use super::{song::Song, song_queue::SongQueue, subprocess::ffmpeg_pcm};
use crate::util::send_embed_http;
use serenity::{
    async_trait,
    client::Context,
    http::Http,
    model::{channel::Message, id::ChannelId},
    prelude::Mutex as SerenityMutex,
};
use songbird::{
    input::{self, reader::Reader},
    tracks::{TrackCommand, TrackHandle},
    Call, Event, EventContext, EventHandler as VoiceEventHandler, TrackEvent,
};
use std::sync::Arc;
use std::time::Duration;
use tokio::sync::Mutex;
use tokio::time::sleep;

pub struct AudioState {
    queue: SongQueue,
    handler: Arc<SerenityMutex<Call>>,
    current_song: Option<Song>,
    track_handle: Option<TrackHandle>,

    channel_id: ChannelId,
    http: Arc<Http>,
}

impl AudioState {
    pub fn new(
        handler: Arc<SerenityMutex<Call>>,
        ctx: &Context,
        msg: &Message,
    ) -> Arc<Mutex<AudioState>> {
        let audio_state = Arc::new(Mutex::new(AudioState {
            queue: SongQueue::new(),
            handler,
            current_song: None,
            track_handle: None,

            channel_id: msg.channel_id,
            http: ctx.http.clone(),
        }));
        let my_audio_state = audio_state.clone();
        tokio::spawn(async move {
            // Leave if no music is playing within 1 minute
            sleep(Duration::from_secs(60)).await;
            let audio_state = my_audio_state.lock().await;
            let current_song = audio_state.current_song.as_ref();
            if current_song.is_none() {
                let mut handler = audio_state.handler.lock().await;
                if let Err(e) = handler.leave().await {
                    println!("Automatic leave failed: {:?}", e);
                }
            }
        });
        audio_state
    }

    pub async fn set_context(self: &mut AudioState, ctx: &Context, msg: &Message) {
        self.channel_id = msg.channel_id;
        self.http = ctx.http.clone();
    }

    pub async fn play_audio(state_mutex: Arc<Mutex<AudioState>>) {
        let mut state = state_mutex.lock().await;
        let song = state.queue.pop().await;

        let song = match song {
            Some(song) => song,
            None => {
                let mut handler = state.handler.lock().await;
                if let Err(e) = handler.leave().await {
                    println!("Error leaving channel: {:?}", e);
                }
                return;
            }
        };

        let source = ffmpeg_pcm(&song.url).await;
        let source = match source {
            Ok(source) => source,
            Err(why) => {
                println!("Error in AudioState::play_audio: {}", why);
                return;
            }
        };
        let reader = Reader::Extension(source);
        let source = input::Input::float_pcm(true, reader);

        let mut handler = state.handler.lock().await;

        let handle = handler.play_source(source);
        drop(handler);

        if let Err(why) = handle.add_event(
            Event::Track(TrackEvent::End),
            SongEndNotifier {
                audio_state: state_mutex.clone(),
            },
        ) {
            panic!("Err AudioState::play_audio: {:?}", why);
        }
        send_embed_http(
            state.channel_id,
            state.http.clone(),
            &format!("Now playing:\n\n {}", song.get_string()),
        )
        .await;
        state.current_song = Some(song);
        state.track_handle = Some(handle);
    }

    pub async fn add_audio(audio_state: Arc<Mutex<AudioState>>, query: &str) {
        let song = match Song::from_query(query).await {
            Ok(song) => song,
            Err(why) => {
                println!("Error add_audio: {}", why);
                return;
            }
        };
        let mut state = audio_state.lock().await;
        state.queue.push(vec![song]).await;
        if state.current_song.is_none() {}
    }

    pub async fn send_track_command(&mut self, cmd: TrackCommand) -> Result<(), String> {
        match self.track_handle.as_ref() {
            Some(th) => match th.send(cmd) {
                Ok(()) => Ok(()),
                Err(why) => Err(format!("{:?}", why)),
            },
            None => Err("no song currently playing".to_string()),
        }
    }

    pub async fn clear(&mut self) -> Result<(), String> {
        self.queue.clear().await
    }

    pub fn get_string(&self) -> String {
        let current_song = match self.current_song.as_ref() {
            Some(song) => song.get_string(),
            None => "*Not playing*\n".to_string(),
        };
        format!(
            "**Current Song:**\n{}\n\n**Queue:**\n{}",
            current_song,
            self.queue.get_string()
        )
    }
}

struct SongEndNotifier {
    audio_state: Arc<Mutex<AudioState>>,
}

#[async_trait]
impl VoiceEventHandler for SongEndNotifier {
    async fn act(&self, _ctx: &EventContext<'_>) -> Option<Event> {
        AudioState::play_audio(self.audio_state.clone()).await;

        None
    }
}
